/*

Program:	MFUUD ver. 3.7.0 (Last Update: 04/12/94)
Author:		Paul Erickson
Affiliation:	Department of Computer Engineering
		Carnegie Mellon University
Email:		erickson+@cmu.edu

**************************************************************

Copyright (c) 1993 Paul Erickson

All rights reserved.


This is the intellectually property of Paul Erickson.  You may use and 
distribute this program freely in either source code or binary form provided 
that the above copywrite notice and this paragraph are duplicated in all such 
forms.  This program cannot be modified without the consent of the author, and 
such modifications include, but are not limited to, incorporating the 
following code into other sources.  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT 
ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT WITHOUT LIMITATION, THE 
IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.

**************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#define CENTURY 1900
#define CYEAR 1994
#define SUBJECT "Subject: "
#define LINELEN 120
#define NAMELEN 80
#define DIRLEN 180
#define SUBJLEN strlen(SUBJECT)
#define TRANS(c) ( (c - ' ') & 077 )
#define MAXKEYW 10
#define MEMLOC (char *) (mem->head + mem->offset)
#define UUCNT 3
#define STRIP(str) str[strlen(str) - 1] = '\0'
#define TLEN 3

struct liststruct {
    char *item;
    int read;
    struct liststruct *next;
};

struct listoflists {
    char *item;
    struct listoflists *next;
    struct liststruct *list;
};

typedef struct liststruct LIST;
typedef struct listoflists LLIST;

struct fliststruct {
    int post_id,num,del,sub_id,maxp;
    long time,seek;
    char *uuname,*filename,*subline;
    LIST *list;
    struct fliststruct *next;
};

typedef struct fliststruct FINFO;

typedef struct {
    char *head;
    int offset;
    int blocks;
    int bsize;
    int avail;
    int items;
}
MEMORY;

char DEL = 0;
char QUIET = 0;
char *DESTDIR = NULL;
int FMODE = 0;
char PLAIN = 0;
char DIR = 0;
char EXT = 0;
char *ERRFILE = NULL;
char *LOGFILE;
char CWD[DIRLEN];
char LOGUPDATE = 0;
char *MARKER = "*";
char DOS = 0;

FILE *fopen();
void GetNum();
int GetId();
void decode();
void SortFiles();
long GetTime();
int uuline();
unsigned int additem();
void meminit();
int GetOption();
char *FormatStr();
void KeySplit();
int keycheck();
char *upcase();
char *userid();
int ReadBline();
void logdir();
void LoadDlist();
int inDlist();
void UpdateDlist();
void AddtoFlist();
void decode_p();
int wc();
void dosname();
void checkname();
int bline();
void GetFileName();

int main(argc,argv)
int argc;
char *argv[];
{

    FILE *file;
    int fcount,aindex,count,keynum,uuchk,header,closed,newline,temp,stop,kchk,n,m;
    long spos,hd_pos;
    char line[LINELEN],subline[LINELEN],name[NAMELEN],*keystr,*keyword[MAXKEYW],
    *cpos,*fn,tstr[NAMELEN],filename[NAMELEN];
    FINFO *flistptr,*filelist,*fptr;
    MEMORY table;
    LLIST *listptr = NULL;
    LIST *list;


    getcwd(CWD,DIRLEN-2);
    flistptr = filelist = NULL;
    meminit(&table,256);
    keynum = stop = 0;
    for (count = 1; count < argc; count++) {
	if ( *argv[count] != '-' ) break;
	switch(GetOption(argv[count])) {
	    case 0:
		fprintf(stderr,"mfuud: ambiguous or unknown command line option \"%s\"\n",argv[count]);
		exit(1);
	    case 1:
		DEL = 1;
		break;
	    case 2:
		QUIET = 1;
		break;
	    case 3:
		EXT = 1;
		break;
	    case 4:
		if (++count < argc) DESTDIR = argv[count];
		break;
	    case 5:
		if (++count < argc) sscanf(argv[count],"%o",&FMODE);
		break;
	    case 6:
		PLAIN = 1;
		break;
	    case 7:
		DIR = 1;
		break;
	    case 8:
		if (++count < argc) keystr = argv[count];
		KeySplit(keyword,keystr,&keynum);
		break;
	    case 9:
		if (++count < argc) ERRFILE = argv[count];
		break;
	    case 10:
		if (++count < argc) logdir(argv[count]);
		stop = 1;
		break;
	    case 11:
		if (++count < argc) {
		    LoadDlist(&listptr);
		    LOGUPDATE = 1;
		    UpdateDlist(listptr,argv[count]);
		}
		stop = 1;
		break;
	    case 12:
		DOS = 1;
		break;
	}
    }
    aindex = count;
    if ( stop ) return(0);
    if ( !QUIET ) printf("Multiple File UUDecoder ver 3.7.0\n");
    if ( (argc-count) == 0 ) {
	decode(NULL);
	return(0);
    }
    if ( PLAIN ) {
	decode_p(&flistptr,count,argc,argv);
	return(0);
    }
    closed = 1;
    LoadDlist(&listptr);
    printf("\n");
    for ( fcount = 0; count < argc;) {
	if ( !EXT || closed ) {
	    spos = header = 0;
	    fn = argv[count];
	    if ( inDlist(listptr,fn,&list) ) {
		count++;
		continue;
	    }
	    if ( (file = fopen(fn,"r")) == NULL ) {
		fprintf(stderr,"mfuud: cannot access file %s\n",argv[count]);
		count++;
		continue;
	    }
	}
	if ( !filelist ) filelist = (FINFO *) malloc ( sizeof(FINFO) );
	filelist->uuname = NULL;
	filelist->del = 0;
	filelist->sub_id = -1;
	filelist->list = list;
	for ( kchk = uuchk = closed = header = 0, newline = 1, *subline = *filename = '\0';
	     (uuchk < UUCNT) && (fgets(line,LINELEN,file)); spos = ftell(file)) {
	    if ( (*line == '\n') && ( header > 5 ) ) filelist->seek = hd_pos;
	    if ( newline )
		if ( (cpos = strchr(line,' ')) && (cpos[-1] == ':') ) {
		    if ( !header ) hd_pos = spos;
		    header++;
		}
		else header = 0;
	    newline = ( line[strlen(line)-1] == '\n' );
	    if ( !strncmp(line,SUBJECT,SUBJLEN) && (*subline == '\0') )
		if ( keycheck(line,keyword,keynum) >= 0 ) {
		    strcpy(subline,line+SUBJLEN);
		    upcase(line);
		    GetFileName(line,filename);
		    GetNum(line,&filelist->num,&filelist->maxp);
		    continue;
		}
		else {
		    kchk = 1;
		    break;
		}
	    if ( (*subline != '\0') && uuline(line) ) uuchk++;
	    if ( !strncmp(line,"Date: ",6) ) filelist->time = GetTime(line+6);
	    if ( !strncmp(line,"From: ",6) ) strcpy(tstr,userid(line+6));
	    if ( (*subline != '\0') && bline(line) ) {
		if ( wc(line) < 3 ) {
		    line[strlen(line)-1] = ' ';
		    fgets(line+strlen(line),LINELEN-strlen(line),file);
		}
		if ( ReadBline(line,&temp,name) ) {
		    for ( fptr = flistptr; fptr; fptr=fptr->next )
			if ( fptr->uuname && !strcmp(name,fptr->uuname) ) {
			    filelist->uuname = fptr->uuname;
			    break;
			}
		    if ( !fptr ) {
			filelist->uuname = (char *) malloc( strlen(name) + 2 );
			strcpy(filelist->uuname,name);
		    }
		}
	    }
	    if ( *filename ) {
		upcase(line);
		if ( strstr(line,filename) && strchr(line,'/') ) {
		    GetNum(line,&n,&m);
		    if ( (n > 0) && (m > 0) ) {
			filelist->num = n;
			if ( (m < filelist->maxp) || (filelist->maxp == -1) ) filelist->maxp = m;
		    }
		}
	    }
	}
	if ( !EXT || feof(file) ) {
	    fclose(file);
	    closed = 1;
	    count++;
	}
	if ( (uuchk >= UUCNT) && (*subline != '\0') ) {
	    filelist->post_id = GetId(&table,tstr);
	    filelist->filename = fn;
	    subline[strlen(subline) - 1] = '\0';
	    if ( !QUIET ) printf("%s\n",subline);
	    upcase(subline);
	    filelist->subline = ( (char *) malloc ( strlen(subline) + 2 ) ) + 1;
	    filelist->subline[-1] = ' ';
	    strcpy(filelist->subline,subline);
	    if ( ( filelist->num == 1 ) && ( filelist->uuname == NULL ) ) filelist->num = -1;
	    if ( (filelist->uuname) && (filelist->num != 1) && (filelist->maxp != -1) ) filelist->uuname = NULL;
	    fcount++;
	    AddtoFlist(&flistptr,filelist);
	    filelist = NULL;
	}
	else if ( !fcount && (*subline == '\0') && !kchk ) {
	    if (!QUIET) printf("Found ordinary uuencoded files.\n");
	    decode_p(&flistptr,aindex,argc,argv);
	    return(0);
	}

    }
    if ( argc == 0 ) {
	fprintf(stderr,"Nothing decoded.\n");
	return(2);
    }
    if ( !QUIET ) printf("\nFound %d uuencoded file segments.\n",fcount);
    if ( !fcount ) return(2);
    SortFiles(&flistptr);
    decode(flistptr);
    printf("\n");
    UpdateDlist(listptr,".:.");
    return(0);
}

void GetNum(string,n,m)
char *string;
int *n,*m;
{
    int aindex, end, brk, start;
    int found = 0;
    char *pos,*temp;

    brk = 0;
    for ( end = start = aindex = 0; string[aindex] != '\0'; aindex++ ) {
	if ( (isdigit(string[aindex])) && (end != -1) ) {
	    start = aindex;
	    end = -1;
	}
	if ( (!isdigit(string[aindex])) && (end == -1) ) end = aindex;
	if ( ((string[aindex] == '/') || brk || (strncmp(string+aindex,"OF",2) == 0)) && (end > start) ) {
	    aindex = -1;
	    break;
	}
	if ( !strncmp(string+aindex,"PART",4) ) found = 1;
	if ( (!isdigit(string[aindex])) && (end > start) && (string[aindex] != ' ') )end = 0;
    }
    if ( aindex != -1 ) end = aindex;
    if ( (aindex == -1) || found ) {
	string[end] = '\0';
	*n = atoi(string + start);
	for (pos = string+end+1; ( !isdigit(*pos) ) && ( *pos ); pos++ );
	for (temp = pos; isdigit(*pos); pos++ );
	*pos = '\0';
	if ( *temp ) *m = atoi(temp);
	else *m = -1;
    }
    else *m = *n = -1;

}

int GetId(table,str)
char *str;
MEMORY *table;
{
    int count;
    char *aindex;

    for (aindex = (table->head), count = 1; count <= (table->items); count++) {
	if ( !strcmp(aindex,str) ) return(count);
	aindex += strlen(aindex) + 1;
    }
    additem(table,str);
    return(table->items);
}

void decline(cptr,file)
char *cptr;
FILE *file;
{
    int n = TRANS(*cptr++);

    while ( n > 0 ) {
	if ( n >= 1 ) putc(TRANS(cptr[0]) << 2 | TRANS(cptr[1]) >> 4,file);
	if ( n >= 2 ) putc(TRANS(cptr[1]) << 4 | TRANS(cptr[2]) >> 2,file);
	if ( n >= 3 ) putc(TRANS(cptr[2]) << 6 | TRANS(cptr[3]),file);
	cptr += 4;
	n -= 3;
    }
}

long maxseek(ptr1, ptr2, spos)
FINFO *ptr1,*ptr2;
long spos;
{

    for (; ptr1; ptr1 = ptr1->next )
	if ( (ptr1 != ptr2) &&
	    (ptr1->filename == ptr2->filename) &&
	    (ptr1->seek > ptr2->seek) &&
	    ( spos > ptr1->seek ) ) spos = ptr1->seek;
    return(spos);
}

void clean(flistptr)
FINFO *flistptr;
{
    int marker;
    FINFO *ptr1,*ptr2;

    for ( ptr1 = flistptr; ptr1; ptr1 = ptr1->next )
	if ( ptr1->del ) {
	    marker = 1;
	    for ( ptr2 = flistptr; ptr2; ptr2 = ptr2->next)
		if ( ( ptr1->filename == ptr2->filename ) && (!ptr2->del) ) marker = 0;
	    if ( marker ) {
		if ( DEL ) unlink(ptr1->filename);
		if ( ptr1->list ) ptr1->list->read = 1;
	    }
	}
}

FILE *GetFilePtr(filename)
char *filename;
{
    FILE *fileptr;

    fileptr = NULL;
    if ( filename ) {
	fileptr = fopen(filename,"a");
	if ( !fileptr ) fprintf(stderr,"%s: cannot open\n",filename);
    }
    return(fileptr);
}

void decode(flistptr)
FINFO *flistptr;
{
    char *CreatePath();
    void BaseName();

    FILE *infile,*outfile,*efile;
    char line1[LINELEN],line2[LINELEN],line[LINELEN],name[NAMELEN],*path;
    int begin,mode;
    long max;
    FINFO *filelist,*fptr;

    begin = 0;
    *line1 = '\0';
    *line2 = '\0';

    printf("\n");
    if ( DOS ) for (fptr = flistptr; fptr; fptr = fptr->next ) dosname(fptr->uuname);
    filelist = flistptr;
    efile = GetFilePtr(ERRFILE);
    do {
	if ( filelist ) {
	    if ( filelist->sub_id == 0 ) {
		filelist->del = 1;
		clean(flistptr);
		continue;
	    }
	    if ( (infile = fopen(filelist->filename,"r")) == NULL ) {
		fprintf(stderr,"mfuud: cannot access file %s\n",filelist->filename);
		exit(1);
	    }
	    fseek(infile,0,SEEK_END);
	    max = maxseek(flistptr,filelist,ftell(infile));
	    fseek(infile,filelist->seek,SEEK_SET);
	}
	else infile = stdin;
	while ( ( (infile == stdin) ? 1 : ftell(infile) < max ) && fgets(line,LINELEN,infile)  )
	    if ( filelist && (filelist->sub_id == -1) ) {
		if ( efile ) fputs(line,efile);
	    }
	    else {
		if ( bline(line) && ( ( !filelist ) || ( (filelist->uuname) && (strcmp(filelist->uuname,MARKER)))) ) {
		    if ( wc(line) < 3 ) {
			line[strlen(line)-1] = ' ';
			fgets(line+strlen(line),LINELEN-strlen(line),infile);
		    }
		    ReadBline(line,&mode,name);
		    if ( begin ) {
			if (!QUIET) printf("Short file\n");
			fclose(outfile);
		    }
		    if ( !DIR ) BaseName(name);
		    if ( DOS ) checkname(flistptr,name);
		    if ( (outfile = fopen(path = CreatePath(name),"wb")) == NULL ) {
			fprintf(stderr,"mfuud: unable to create %s\n",path);
			break;
		    }
		    begin = 1;
		    if ( filelist && (*(filelist->uuname) != '\\') ) strcpy(filelist->uuname,MARKER);
		    if (!QUIET) printf("UUdecoding %s\n",name);
		    if ( FMODE ) chmod(path,FMODE);
		    else chmod(path,mode);
		}
		if ( begin && (uuline(line)) ) decline(line,outfile);
		if ( (strncmp(line,"end",3) == 0) && begin ) {
		    decline(line1,outfile);
		    decline(line2,outfile);
		    begin = 0;
		    fclose(outfile);
		}
		strcpy(line1,line2);
		strcpy(line2,line);
	    }
	if ( filelist && ( (filelist->sub_id != -1) || efile )) {
	    fclose(infile);
	    filelist->del = 1;
	    clean(flistptr);
	}
    }
    while ( filelist && ( filelist = filelist->next ) );
    if ( begin ) {
	if (!QUIET) printf("Short file\n");
	fclose(outfile);
    }
    if ( efile ) fclose(efile);
}

char *CreatePath(name)
char *name;
{
    static char pathname[DIRLEN];

    if ( !DESTDIR ) return(name);
    strcpy(pathname,DESTDIR);
    strcat(pathname,"/");
    strcat(pathname,name);
    return(pathname);
}

char *FormatStr(str)
char *str;
{
    int aindex;
    for (aindex = 0; str[aindex] != '\0'; aindex++)
	if ( (isalnum(str[aindex]))
	    || (str[aindex] == '+')
	    || (str[aindex] == '-') ) str[aindex] = toupper(str[aindex]);
	else str[aindex] = ' ';
    return(str);
}

long GetTime(str)
char *str;
{
    char *months = "aJanbFebcMardApreMayfJungJulhAugiSepjOctkNovlDec";
    char temp[10];
    char month[10],*pos;
    long day,hours,min,sec,mnum,year;

    if ( sscanf(str,"%s %ld %s %ld %ld:%ld:%ld",temp,&day,month,&year,&hours,&min,&sec) != 7)
	if ( sscanf(str,"%ld %s %ld %ld:%ld:%ld",&day,month,&year,&hours,&min,&sec) != 6 )
	    return(0);
    pos = strstr(months,month);
    if ( pos ) mnum = pos[-1] - 'a';
    else return(0);
    if ( year < 100 ) year += CENTURY;
    year -= CYEAR;
    return (year * 32140800 + day * 86400 + mnum * 2678400 + hours * 3600 + min * 60 + sec );
}

int uuline(line)
char *line;
{
    if ( (*line == 'M') && ( (strlen(line)==62) || (strlen(line)==63) ) ) return(1);
    else return(0);
}

void meminit(mem,bsize)
MEMORY *mem;
int bsize;
{
    mem->head = NULL;
    mem->offset = 0;
    mem->blocks = 0;
    mem->bsize = bsize;
    mem->avail = 0;
    mem->items = 0;
}

unsigned int additem(mem, item)
MEMORY *mem;
char *item;
{
    unsigned int retvalue;
    int size;

    if (( mem->avail ) < ( strlen(item) + 1 )) {
	(mem->blocks)++;
	size = (mem->blocks) * (mem->bsize);
	mem->avail += mem->bsize;
	if (mem->head)
	    mem->head = (char*) realloc(mem->head,size);
	else
	    mem->head = (char*) malloc(size);
    }
    strcpy(MEMLOC,item);
    retvalue = mem->offset;
    mem->offset += strlen(item) + 1;
    mem->avail -= strlen(item) + 1;
    (mem->items)++;
    return(retvalue);
}

int GetOption(str)
char *str;
{
    int i,match;
    char *ptr;
    char *optlist = "-delete\0-quiet\0-extended\0-dest\0-mode\0-plain\0-dir\0-keywords\0-errors\0-log\0-unlog\0-dos\0";

    for ( i = 1, match = 0, ptr = optlist; *ptr != '\0'; i++, ptr += strlen(ptr) + 1 ) {
	if ( !strncmp(ptr,str,strlen(str)) )
	    if ( match ) return(0);
	    else match = i;
    }
    return(match);
}

void BaseName(path)
char *path;
{
    char *cpos,*lpos,temp[NAMELEN];

    lpos = path;
    cpos = path;

    for (; cpos; ) if ( cpos = strchr(cpos + 1,'/') ) lpos = cpos + 1;
    strcpy(temp,lpos);
    strcpy(path,temp);
}

void KeySplit(strarr,str,keynum)
char **strarr,*str;
int *keynum;
{
    char *ptr;

    ptr = strtok(str," ");
    while (ptr) {
	strarr[(*keynum)++] = FormatStr(ptr);
	ptr = strtok(NULL," ");
    }
}

int keycheck(str,keyword,keynum)
char *str,**keyword;
int keynum;
{
    char tstr[LINELEN];
    int aindex;
    int value = 0;

    if ( !keynum ) return(keynum);
    strcpy(tstr,str);
    FormatStr(tstr);

    for (aindex = 0; aindex < keynum; aindex++)
	if (strstr(tstr,keyword[aindex]+1))
	    value += ( (*(keyword[aindex]) == '-') ? -1 : 1 );
    return(value);
}

char *upcase(str)
char *str;
{
    int aindex;
    for (aindex = 0; str[aindex] != '\0'; aindex++) str[aindex] = toupper(str[aindex]);
    return(str);
}

int same(s1,s2)
char *s1,*s2;
{
    int cont = 0;

    while (*s1 && *s2) {
	if ( !isalnum(*s1) ) {
	    s1++;
	    cont = 1;
	}
	if ( !isalnum(*s2) ) {
	    s2++;
	    cont = 1;
	}
	if ( cont ) {
	    cont = 0;
	    continue;
	}
	if ( *s1 != *s2 ) return(0);
	s1++;
	s2++;
    }
    return(*s1 == *s2);
}

void GetFileName(line,nstr)
char *line,*nstr;
{
    int aindex, start, end, period;
    char *temp,string[LINELEN];
    strcpy(string,line);
    for (start = end = period = aindex = 0; string[aindex] != '\0'; aindex++ ) {
	if ( (string[aindex] != ' ') && (end != -1) ) {
	    start = aindex;
	    end = -1;
	}
	if ( (string[aindex] == '.') && (string[aindex+1] != '.') && (string[aindex-1] != '.') ) period = 1;
	if ( ((string[aindex] == ' ') || (string[aindex] == '(')) && (end == -1 ) ) {
	    end = aindex;
	    if ( period == 1 ) break;
	}
    }
    if ( period == 1 ) {
	temp = strtok(string+start," ,()\n");
	strcpy(nstr,temp);
    }
    else *nstr = '\0';
}
void SortFiles(flistptr)
FINFO **flistptr;
{
    int result,match,bmatch,ncount,lnum,id,pos,len,idnum,dcount,maxp,idmatch,pass;
    char idstr[LINELEN],pstr[LINELEN],*fname;
    FINFO *start,*fptr,*fptr2,**lptr,**lptr2,*temp;

    for ( fptr = *flistptr; fptr; fptr = fptr->next )
	if ( fptr->del ) {
	    fptr->subline = upcase(fptr->uuname);
	    fptr->del = 0;
	}
    for ( idnum = 1, fptr = *flistptr; fptr; ) {
	id = fptr->post_id;
	start = fptr;
	for ( pass = 0; pass < 2; pass++ )
	    for (fptr = start; fptr && ( fptr->post_id == id ); fptr=fptr->next) {
		if ( (fptr->sub_id) != -1 ) continue;
		idmatch = 0;
		match = 0;
		pos = -1;
		do {
		    switch(pass) {
			case 0:
			    GetFileName(fptr->subline,pstr);
			    pos = strlen(fptr->subline);
			    break;
			case 1:
			    if ( match && ( (pos+len-1) < strlen(fptr->subline)) ) len++;
			    else {
				pos++;
				len = 3;
			    }
			    strncpy(pstr,fptr->subline + pos,len);
			    pstr[len] = '\0';
			    break;
		    }
		    if ( *pstr == '\0' ) break;
		    lnum = 0;
		    fname = NULL;
		    for ( fptr2 = start, match = dcount = bmatch = ncount = 0, maxp = -1; fptr2 && ( fptr2->post_id == id ); fptr2=fptr2->next ) {
			if ( fptr2->sub_id != -1 ) continue;
			if ( strstr(fptr2->subline,pstr) ) {
			    match++;
			    if ( (fptr2->uuname != NULL) && (fptr2->uuname != fname) ) {
				bmatch++;
				fname = fptr2->uuname;
			    }
			    if ( fptr2->maxp != -1 ) maxp = fptr2->maxp;
			    if ( fptr2->num == -1 ) dcount++;
			    else if ( lnum != fptr2->num ) {
				ncount += ( fptr2->num - lnum - 1 );
				lnum = fptr2->num;
			    }
			    else if ( strstr(fptr2->subline,"REPOST") ) match--;
			    else ncount++;
			}
		    }
		    if ( (match > idmatch) && ( bmatch == 1 ) && (ncount <= dcount) && (maxp != -1 ? match == maxp : 1) ) {
			strcpy(idstr,pstr);
			idmatch = match;
		    }
		    if ( idmatch && ( match == maxp ) ) break;
		}
		while (pos < ( strlen(fptr->subline) - TLEN + 1) );
		if ( idmatch ) {
		    for ( fptr2 = start; fptr2 && (fptr2->post_id == id); fptr2=fptr2->next ) {
			if ( fptr2->sub_id != -1 ) continue;
			if ( strstr(fptr2->subline,idstr) ) fptr2->sub_id = idnum;
		    }
		    idnum++;
		}
	    }
	for ( fptr2 = start, lnum = match = bmatch = 0, maxp = -1; fptr2 && (fptr2->post_id == id); fptr2=fptr2->next ) {
	    if ( (fptr2->sub_id == -1) && (lnum + 1 == fptr2->num) ) match++;
	    else break;
	    if ( fptr2->maxp != -1 ) maxp = fptr2->maxp;
	    if ( fptr2->uuname ) bmatch++;
	    lnum = fptr2->num;
	}
	if ( (match == maxp) && (bmatch == 1) ) {
	    for ( fptr2 = start; fptr2 && (fptr2->post_id == id); fptr2=fptr2->next )
		if ( fptr2->sub_id == -1 ) fptr2->sub_id = idnum;
	    idnum++;
	}
    }
    if ( !QUIET ) for ( match = 1, fptr = *flistptr; fptr; fptr=fptr->next )
	if ( fptr->sub_id == -1 ) {
	    if ( match ) {
		printf("\nCaptions for (file) segments not decoded:\n");
		match = 0;
	    }
	    printf("%s\n",fptr->subline);
	}
    if ( !match ) printf("\n");
    lptr2 = lptr = flistptr;
    do {
	match = 0;
	id = (*lptr)->post_id;
	for (; (*lptr)->next && ( ((*lptr)->next)->post_id == id ); lptr = &((*lptr)->next) ) {
	    if ( (result = (*lptr)->post_id - ((*lptr)->next)->post_id) == 0 ) {
		if ( (result = (*lptr)->sub_id - ((*lptr)->next)->sub_id) == 0 ) {
		    if ( ( ((*lptr)->next)->num == (*lptr)->num ) && ( ((*lptr)->next)->num != -1 ) ) {
			if ( strstr(((*lptr)->next)->subline,"REPOST") ) (*lptr)->sub_id = 0;
			if ( strstr((*lptr)->subline,"REPOST") ) ((*lptr)->next)->sub_id = 0;
		    }
		    if ( ( (*lptr)->num == -1 ) || ( ((*lptr)->next)->num == -1 ) )
			result = (*lptr)->time - ((*lptr)->next)->time;
		    else result = (*lptr)->num - ((*lptr)->next)->num;
		}
	    }
	    if ( result > 0 ) {
		temp = *lptr;
		*lptr = (*lptr)->next;
		temp->next = temp->next->next;
		(*lptr)->next = temp;
		match = 1;
	    }
	}
	lptr = &(*lptr)->next;
	if ( match ) lptr = lptr2;
	else lptr2 = lptr;
    }
    while ( *lptr );
    printf("%d complete (non-segmented) files found.\n",idnum-1);
}

char *userid(str)
char *str;
{
    char *token,*pos;
    char *cset = "() <>,\n";

    while ( (token = strtok(str,cset)) != NULL ) {
	if (str) str = NULL;
	if ( pos = strchr(token,'@') ) {
	    *pos = '\0';
	    return(token);
	}
    }
    return("NONE!");
}

LLIST *getptrptr(ptrptr,item)
LLIST **ptrptr;
char *item;
{
    for (; *ptrptr; ptrptr = &((*ptrptr)->next) )
	if ( !strcmp((*ptrptr)->item, item) ) return(*ptrptr);
    *ptrptr = (LLIST *) malloc ( sizeof(LLIST) );
    (*ptrptr)->item = (char *) malloc( strlen(item) + 1);
    strcpy((*ptrptr)->item,item);
    (*ptrptr)->next = NULL;
    (*ptrptr)->list = NULL;
    return(*ptrptr);
}

LIST *addtolist(ptr,item,dir)
LIST *ptr;
char *item;
char *dir;
{
    LIST *temp_ptr;

    temp_ptr = (LIST *) malloc ( sizeof(LIST) );
    temp_ptr->item = (char *) malloc ( strlen(item) + 1);
    strcpy(temp_ptr->item,item);
    temp_ptr->next = ptr;
    if ( dir && strcmp(CWD,dir) ) temp_ptr->read = 1; else temp_ptr->read = 0;
    return(temp_ptr);
}

void LoadDlist(llistptrptr)
LLIST **llistptrptr;
{
    FILE *file;
    char line[DIRLEN];
    LLIST *listptr;

    LOGFILE = getenv("MFUUDLOG");
    if ( !LOGFILE ) return;
    if ( (file = fopen(LOGFILE,"r")) == NULL ) return;
    if (!QUIET) printf("Reading %s\n",LOGFILE);
    while ( fgets(line,DIRLEN,file) ) {
	STRIP(line);
	if ( *line == '*' ) listptr = getptrptr(llistptrptr,line+1);
	if ( *line == '@' ) listptr->list = addtolist(listptr->list,line+1,listptr->item);
    }
    fclose(file);
}

int ReadBline(line,mode,filename)
char *line,*filename;
int *mode;
{
    int result;

    for (; isspace(*line); ++line);
    result = sscanf(line, "begin %o %s",mode,filename);
    if (  result != 2 ) {
	result = sscanf(line, "begin %s",filename);
	*mode = 644;
    }
    return(result);
}

void logdir(dirname)
char *dirname;
{
    FILE *file;

    LOGFILE = getenv("MFUUDLOG");
    if ( !LOGFILE ) {
	fprintf(stderr,"mfuud: MFUUDLOG enviroment variable not set\n");
	return;
    }
    if ( (file = fopen(LOGFILE,"a")) == NULL ) {
	fprintf(stderr,"mfuud: can't open log file\n");
	return;
    }
    fprintf(file,"*%s\n",dirname);
    fclose(file);
}

char *MakePath(filename)
char *filename;
{
    static char tempstr[DIRLEN];
    char *ptr1,*ptr2;

    if ( *filename == '/' ) {
	strcpy(tempstr,filename);
	return ( tempstr );
    }
    strcpy(tempstr,CWD);
    for (ptr1 = filename, ptr2 = tempstr + strlen(tempstr); !strncmp(ptr1,"../",3); ptr1 += 3) {
	for (; ( *ptr2 != '/' ) && (ptr2 != tempstr); ptr2--);
	*ptr2 = '\0';
    }
    strcat(tempstr,"/");
    strcat(tempstr,ptr1);
    return( tempstr );
}

int inDlist(listptr,fn,list)
LLIST *listptr;
char *fn;
LIST **list;
{
    LLIST *ptr;
    LIST *ptr2;
    char *item;
    int result = 0;

    item = MakePath(fn);
    *list = NULL;
    for (ptr = listptr; ptr; ptr = ptr->next)
	if ( !strncmp(item,ptr->item,strlen(ptr->item)) ) {
	    LOGUPDATE = 1;
	    for (fn = item + strlen(item); ( *fn != '/' ) && (fn != item); fn--);
	    *(fn++) = '\0';
	    ptr = getptrptr(&listptr,item);
	    for ( ptr2 = ptr->list; ptr2; ptr2=ptr2->next )
		if ( !strcmp(ptr2->item,fn) ) {
		    ptr2->read = result = 1;
		    break;
		}
	    if ( !ptr2 ) {
		ptr->list = addtolist(ptr->list,fn,NULL);
		*list = ptr->list;
	    }
	    break;
	}
    return(result);
}

void UpdateDlist(listptr,xitem)
LLIST *listptr;
char *xitem;
{
    FILE *file;
    LLIST *lptr;
    LIST *ptr;

    if ( !LOGFILE || !LOGUPDATE) return;
    if ( (file = fopen(LOGFILE,"w")) == NULL ) {
	fprintf(stderr,"mfuud: can't open log file\n");
	return;
    }
    if (!QUIET) printf("Updating %s\n",LOGFILE);
    for (lptr = listptr; lptr; lptr = lptr->next)
	if ( strncmp(lptr->item,xitem,strlen(xitem)) ){
	    fprintf(file,"*%s\n",lptr->item);
	    for (ptr = lptr->list; ptr; ptr = ptr->next)
		if ( ptr->read ) fprintf(file,"@%s\n",ptr->item);
	}
    fclose(file);
}

void AddtoFlist(list,node)
FINFO **list,*node;
{
    FINFO *ptr,**nptr;
    int result;

    for (ptr = *list, nptr = list; ptr; nptr= &(ptr->next), ptr = ptr->next) {
	if ( (result = (ptr->post_id - node->post_id) ) == 0 )
	    result = (ptr->num - node->num);
	if ( !result && same(ptr->subline,node->subline) )
	    if ( ptr->uuname == node->uuname ) {
		if ( ptr->time > node->time ) node->sub_id = 0;
		else ptr->sub_id = 0;
	    }
	    else {
		if ( ptr->uuname ) ptr->del = 1;
		if ( node->uuname ) node->del = 1;
	    }
	if ( result > 0 ) break;
    }
    *nptr = node;
    node->next = ptr;
}

void decode_p(flistptr,index,argc,argv)
FINFO **flistptr;
char **argv;
int argc,index;
{
    FINFO **lptr;
    FINFO *filelist;

    for (lptr = flistptr; index < argc; index++, lptr = &((*lptr)->next)) {
	filelist = (FINFO *) malloc ( sizeof(FINFO) );
	filelist->filename = argv[index];
	filelist->del = 0;
	filelist->sub_id = 1;
	filelist->seek = 0;
	filelist->list = NULL;
	filelist->uuname = "\\";
	*lptr = filelist;
    }
    *lptr = NULL;
    decode(*flistptr);
}

int wc(str)
char *str;
{
    int count = 0;
    char temp[LINELEN];

    strcpy(temp,str);
    if ( strtok(temp," \n") ) ++count;

    while ( strtok(NULL," \n") ) ++count;
    return(count);
}

void dosname(name)
char *name;
{
    int i,j;
    static l[3] = {
	6,2,4	};
	char *ptr1,*ptr2;

	if ( !name ) return;
	ptr1 = ptr2 = name;
	for ( i = j = 0; *ptr1; ++ptr1, ++i ) {
	    if ( (!j && (ptr1[2] == '.')) || (*ptr1 == '.') )
		if ( j < 2 ) { i = 0; ++j; }
	    if ( i < l[j] ) *(ptr2++) = *ptr1;
	}
	*(ptr2++) = '\0';
	*ptr2 = '\0';
}

void checkname(filelist,name)
FINFO *filelist;
char *name;
{
    char *nptr,temp[LINELEN],*ext;

    dosname(name);
    for (; filelist; filelist = filelist->next )
	if ( (filelist->uuname) && (!strcmp(filelist->uuname,name)) ) {
	    nptr = filelist->uuname + ( strlen(filelist->uuname) +1 );
	    if ( *nptr ) {
		ext = strchr(name,'.');
		if ( ext ) *(ext++) = '\0';
		else ext = name + strlen(name);
		if ( strlen(name) > 6 ) name[6] = '\0';
		sprintf(temp,"%s%02d.%s",name,(int) *nptr,ext);
		strcpy(name,temp);
	    }
	    ++(*nptr);
	}
}

int bline(ptr)
char *ptr;
{
    for (; isspace(*ptr); ++ptr);
    return ( !strncmp(ptr,"begin ",6) );
}
